// Example HPC++ Program
//
// $Id: hpcxx_test7.C,v 1.1.1.1 1998/05/08 01:04:09 jvillaci Exp $

// Sample server code.
// This program creates to instances of hpcxx_testb.C on two machines.
// It then ships a sparse array to the first. and then the first ships
// it to the second.

#include <stdlib.h>
#include <unistd.h>
#include <stream.h>
#include <signal.h>

#include <hpcxx.h>

#define ElapseTime(Start, Stop) \
    (  (  ( ((double)(Stop.tv_sec)) - ((double)(Start.tv_sec)) ) \
        + (((double)(Stop.tv_usec))/1000000.0) ) \
     - (((double)(Start.tv_usec))/1000000.0) )

// sparse_array is the class that describes the sparse array.

class sparse_array{
 public:
   int *index1;
   int *index2;
   double *vals;
   int size;
   sparse_array(){
      size = 0;
      index1 = index2 = NULL;
      vals = NULL;
      }
   sparse_array(int n){
      index1 = new int[n];
      index2 = new int[n];
      vals = new double[n];
      size = n;
      }
   sparse_array(sparse_array &a){
      size = a.size;
      index1 = a.index1;
      index2 = a.index2;
      vals = a.vals;
     }     
   sparse_array &operator=(sparse_array &a){
      size = a.size;
      index1 = a.index1;
      index2 = a.index2;
      vals = a.vals;
     return *this;
    }
   void init(){
      for(int i = 0; i < size; i++){
          index1[i] = i;
          index2[i] = 10*i;
          vals[i] = 10*i+3.14;
         }
      }
};


// in addition to the class sparse array, we need to write
// the functions which tell how to pack and unpack an array
// of sparse arrays.

void hpcxx_pack(HPCxx_Buffer &b, sparse_array *spa, int size){
     for(int i = 0; i < size; i++){
         hpcxx_pack(b, &spa->size, 1);
         hpcxx_pack(b, spa->vals, spa->size);
         hpcxx_pack(b, spa->index1, spa->size);
         hpcxx_pack(b, spa->index2, spa->size);
     }
}

void hpcxx_unpack(HPCxx_Buffer &b, sparse_array *spa, int size){
     for(int i = 0; i < size; i++){
         hpcxx_unpack(b, &spa->size, 1);
         spa->index1 = new int[spa->size];
         spa->index2 = new int[spa->size];
         spa->vals = new double[spa->size];
         hpcxx_unpack(b, spa->vals, spa->size);
         hpcxx_unpack(b, spa->index1, spa->size);
         hpcxx_unpack(b, spa->index2, spa->size);
     }
}

// the following are the "stub" versions of the functions that
// are called on the hpcxx_testb.C as remote calls.
// we need the stubs so that the functions can be registered
// with the correct type signatures.

// function stores the sparse array.
int deposit(sparse_array sa){
  // stub version
     return 1;
}
int checksa(int n){
  // stub version
  return 0;
}


// function to kill a remote object

int kill_me(int i){
  // stub version
   return 0;
}

// function to send a contextId to a remote object
int setoutid(HPCxx_ContextID id){
   // stub version
   return 0;
}

// function to tell p1 to send a sparse array to p2.
int sendToRemote(int i){
 // stub version
  return 0;
}

// the path to the hpcxx_testb.C executables on p1 and p2.
// you could make them different places if you liked, just #define another
// PATH variable of some sort.

#define PATH "/u/btemko/gp2/hpcxx_slave"
       
int main(int argc, char **argv)
{
  timeval start_time, stop_time;
  float work_time, speed, sar_size;

  HPCxx_Group *group;
  //hpcxx_id_t fetch_id = hpcxx_register(fetch_int, 22);
  hpcxx_id_t kill_id = hpcxx_register(kill_me,23);
  hpcxx_id_t deposit_id = hpcxx_register(deposit, 25);
  hpcxx_id_t checksa_id = hpcxx_register(checksa, 26);
  hpcxx_id_t setoutid_id = hpcxx_register(setoutid, 27);
  hpcxx_id_t sendToRemote_id = hpcxx_register(sendToRemote, 28);

  hpcxx_init(argc, argv, group);

  // create the remote objects.
  // the createContext function returns the contextIDs needed to do the
  // remote function calls.

  HPCxx_ContextID *p1 = hpcxx_createContext("bread", PATH);
  HPCxx_ContextID *p2  = hpcxx_createContext("bread", PATH);


  // tell p1 the hpcxx_contextID for p2.

  int ret_val;
  hpcxx_invoke(*p1, ret_val, setoutid_id, *p2); 

  // create a sparse array.

  int n = 100000;
  sparse_array sar(n);
  sar.init();
  sar_size = n*(2.0*sizeof(int)+sizeof(double) + 32);

  // ship the array to p1.  time it too.
  int z;
  hpcxx_invoke(*p1, z, deposit_id, sar);


  // ask p1 if the array is correct.

  hpcxx_invoke(*p1, z, checksa_id, n);
  printf("[a] errors from p1 = %d\n", z);

  // now tell p1 to send the array to p2.

  hpcxx_invoke(*p1, z, sendToRemote_id, n);

  // ask p2 if it got there o.k.

  hpcxx_invoke(*p2, z, checksa_id, n);
  printf("[a] errors from p2 = %d\n", z);
  
  int i = 0;
  hpcxx_invoke(*p1, z, kill_id, i);
  hpcxx_invoke(*p2, z, kill_id, i);
  //hpcxx_exit(group);
  _exit(0);
}
