// Example HPC++ Program
//
// $Id: hpcxx_testb.C,v 1.1.1.1 1998/05/08 01:04:09 jvillaci Exp $

#include <stdlib.h>
#include <unistd.h>
#include <stream.h>
#include <signal.h>

#include <hpcxx.h>

#define ElapseTime(Start, Stop) \
    (  (  ( ((double)(Stop.tv_sec)) - ((double)(Start.tv_sec)) ) \
        + (((double)(Stop.tv_usec))/1000000.0) ) \
     - (((double)(Start.tv_usec))/1000000.0) )


class sparse_array{
 public:
   int *index1;
   int *index2;
   double *vals;
   int size;
   sparse_array(){
      size = 0; 
      index1 = index2 = NULL;
      vals = NULL;
      }
   sparse_array(int n){
      index1 = new int[n];
      index2 = new int[n];
      vals = new double[n];
      size = n;
      }
   sparse_array &operator=(sparse_array &a){
      size = a.size;
      index1 = a.index1;
      index2 = a.index2;
      vals = a.vals;
     return *this;
    }
   void init(){
      for(int i = 0; i < size; i++){
          index1[i] = i;
          index2[i] = 10*i;
          vals[i] = 10*i+3.14;
         }
      }
};

void hpcxx_pack(HPCxx_Buffer &b, sparse_array *spa, int size){
     for(int i = 0; i < size; i++){
         hpcxx_pack(b, &spa->size, 1);
         hpcxx_pack(b, spa->vals, spa->size);
         hpcxx_pack(b, spa->index1, spa->size);
         hpcxx_pack(b, spa->index2, spa->size);
     }
}

void hpcxx_unpack(HPCxx_Buffer &b, sparse_array *spa, int size){
     for(int i = 0; i < size; i++){
         hpcxx_unpack(b, &spa->size, 1);
         spa->index1 = new int[spa->size];
         spa->index2 = new int[spa->size];
         spa->vals = new double[spa->size];
         hpcxx_unpack(b, spa->vals, spa->size);
         hpcxx_unpack(b, spa->index1, spa->size);
         hpcxx_unpack(b, spa->index2, spa->size);
     }
}

sparse_array spa(0);
sparse_array spa2(100000);

int deposit(sparse_array sa){
     spa = sa;
     return 1;
}
int checksa(int n){
   int count = 0;
   spa2.init();
   for(int i = 0; i < spa.size; i++){
      if(spa.index1[i] != spa2.index1[i]) count++;
      if(spa.index2[i] != spa2.index2[i]) count++;
      if(spa.vals[i] != spa2.vals[i]) count++;
   }
  printf("error count is %d\n", count);
  return count;
}

HPCxx_Sync<int> x;

int kill_me(int i){
   printf("[b] i am dead\n");
   fflush(NULL);
   x = i;
   return 0;
}

HPCxx_GlobalPtr<int> fetch_int(int i){
     printf("[b] starting fetch_int\n");
     int *p = new int[i];
     for(int j = 0; j < i; j++) p[j] = 100+j;
     HPCxx_GlobalPtr<int> gp(p);
     printf("[b] fetch_int called\n");
     return gp;
}


HPCxx_ContextID out_id;

int setoutid(HPCxx_ContextID id){
   out_id = id;
   return 0;
}

hpcxx_id_t deposit_id;
int sendToTwin(int i){
  timeval start_time, stop_time;
  float work_time, speed, sar_size;
  sar_size = 100000*(2.0*sizeof(int)+sizeof(double) + 32);
  int z = 0;
  gettimeofday(&start_time, NULL);
     hpcxx_invoke(out_id, z, deposit_id, spa);
  gettimeofday(&stop_time, NULL);
  work_time = ElapseTime(start_time, stop_time);
  speed = 8*sar_size/(work_time*1.0e6);  // 8 for bits per sec.
  printf("speed for array of size %f MB is %f Mbps\n", sar_size/1.0e6, speed);

  return z;
}
       
int main(int argc, char **argv)
{

  HPCxx_Group *group;
  hpcxx_id_t kill_id  = hpcxx_register(kill_me, 23);
  hpcxx_id_t fetch_id = hpcxx_register(fetch_int, 22);
             deposit_id = hpcxx_register(deposit, 25);
  hpcxx_id_t checksa_id = hpcxx_register(checksa, 26);
  hpcxx_id_t setoutid_id = hpcxx_register(setoutid, 27);
  hpcxx_id_t sendToTwin_id = hpcxx_register(sendToTwin, 28);

  hpcxx_initAgent(argc, argv);
  printf("agent b initialized\n");
  // now wait to be killed
  int i;
  i = x;
  printf("b will exit now\n");
  fflush(NULL);
  return 0;
}
